
#include "include/settings.h"
extern "C"
{
  #include <audacious/plugin.h>
}

#define CUBE_CONFIG_TAG "cube"

static ConfigDb *GetConfigFile( void )
{
  ConfigDb *cfg = aud_cfg_db_open();
  return cfg;
}

void DefaultSettings(LPSETTINGS pSettings)
{
  memset(pSettings,0,sizeof(SETTINGS));
  
  pSettings->looptimes = 2;
  pSettings->fadelength = 0;
  pSettings->fadedelay = 0;
  pSettings->ADXCutoff = 500;
  pSettings->ADXChannel = 0;
  pSettings->rsf_coefficient[0] = 29000;
  pSettings->rsf_coefficient[1] = 13000;
}

bool LoadSettings(LPSETTINGS pSettings)
{
  ConfigDb *cfg = GetConfigFile();
  if (!cfg)
  {
    DefaultSettings(pSettings);
  }
  else
  {  
    bool bRet = (aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"looptimes",&pSettings->looptimes) && 
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"fadelength",&pSettings->fadelength) &&
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"fadedelay",&pSettings->fadedelay) &&
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"adxonechan",(int*)&pSettings->ADXChannel) &&
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"adxcutoff",(int*)&pSettings->ADXCutoff) &&
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"rsfcoef1",(int*)pSettings->rsf_coefficient+0) &&
		 aud_cfg_db_get_int(cfg,CUBE_CONFIG_TAG,"rsfcoef2",(int*)pSettings->rsf_coefficient+1));
    
    aud_cfg_db_close(cfg); 

    // check if reading one value failed.  If so, then use defaults
    if (!bRet)
      DefaultSettings(pSettings);
  }
  return true;
}

bool SaveSettings(LPSETTINGS pSettings)
{
  ConfigDb *cfg = GetConfigFile();
  if (!cfg)
    return false;

  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"looptimes",pSettings->looptimes);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"fadelength",pSettings->fadelength);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"fadedelay",pSettings->fadedelay);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"adxonechan",pSettings->ADXChannel);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"adxcutoff",pSettings->ADXCutoff);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"rsfcoef1",pSettings->rsf_coefficient[0]);
  aud_cfg_db_set_int(cfg,CUBE_CONFIG_TAG,"rsfcoef2",pSettings->rsf_coefficient[1]);
  aud_cfg_db_close(cfg);
  return true;
}
